(in-package :text-utils)

(defun codepoint->int (string)
  (parse-integer string :radix 16))

(defun split-fields (line)
  (cl-ppcre:split ";" line))

(defun valid-line-p (line)
  (and (string-not-empty-p line)
       (not (cl-ppcre:scan "^#" line))))

(defun match-codepoint-p (a b)
  (every #'char= a b))

(defun extract-codepoints (stream)
  "Extract shotcodes from the emoji table file"
  (flet ((readline ()
           (read-line stream nil nil)))
    (let ((res ()))
      (loop for line = (readline) while line do
        (when (valid-line-p line)
          (let* ((fields            (split-fields line))
                 (codepoints-fields (split-words (alexandria:first-elt fields)))
                 (rangesp           (cl-ppcre:scan "\\.\\."
                                                   (first codepoints-fields)))
                 (codepoints-ranges (and rangesp
                                         (mapcar (lambda (a) (codepoint->int a))
                                                 (cl-ppcre:split "\\.\\."
                                                                 (first codepoints-fields)))))
                 (codepoint-list    (and (not rangesp)
                                         (mapcar (lambda (a) (codepoint->int a))
                                                 codepoints-fields))))
            (if (not rangesp)
                (push (mapcar #'code-char codepoint-list) res)
                (loop for i from (first codepoints-ranges) to (second codepoints-ranges) do
                  (push (list (code-char i)) res))))))
      res)))

(defun extract-emojis-codepoints (uri)
  (let* ((data (drakma:http-request uri)))
    (with-input-from-string (stream data)
      (extract-codepoints stream))))

(defmacro gen-matcher-sequences ()
  `(defun emoji-sequences-p (chars)
     (find-if (lambda (a) (match-codepoint-p chars a))
              ',(extract-emojis-codepoints "https://unicode.org/Public/emoji/13.1/emoji-sequences.txt"))))

(defmacro gen-matcher-zwj-sequences ()
  `(defun emoji-zwj-sequences-p (chars)
     (find-if (lambda (a) (match-codepoint-p chars a))
              ',(extract-emojis-codepoints "https://unicode.org/Public/emoji/13.1/emoji-zwj-sequences.txt"))))

;;;; expansion of (gen-matcher-sequences)

(defun emoji-sequences-p (chars)
  (member-if (lambda (a) (match-codepoint-p chars a))
           '((#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D3 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D3 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D3 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D3 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D3 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D2 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D2 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D2 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D2 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D2 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9BB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9BB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9BB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9BB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9BB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9B6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9B6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9B6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9B6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9B6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9B5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9B5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9B5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9B5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9B5 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F977 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F977 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F977 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F977 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F977 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F936 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F936 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F936 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F936 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F936 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F934 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F934 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F934 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F934 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F934 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F933 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F933 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F933 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F933 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F933 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F932 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F932 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F932 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F932 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F932 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F931 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F931 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F931 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F931 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F931 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F930 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F930 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F930 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F930 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F930 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F91F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F91F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F91F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F91F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F91F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F91E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F91E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F91E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F91E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F91E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F91C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F91C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F91C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F91C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F91C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F91B #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F91B #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F91B #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F91B #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F91B #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F91A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F91A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F91A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F91A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F91A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F919 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F919 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F919 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F919 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F919 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SIGN_OF_THE_HORNS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SIGN_OF_THE_HORNS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SIGN_OF_THE_HORNS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SIGN_OF_THE_HORNS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SIGN_OF_THE_HORNS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F90F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F90F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F90F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F90F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F90F #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F90C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F90C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F90C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F90C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F90C #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SLEEPING_ACCOMMODATION #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SLEEPING_ACCOMMODATION #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SLEEPING_ACCOMMODATION #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SLEEPING_ACCOMMODATION #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SLEEPING_ACCOMMODATION #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BATH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BATH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BATH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BATH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BATH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_WITH_FOLDED_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_WITH_FOLDED_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_WITH_FOLDED_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_WITH_FOLDED_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_WITH_FOLDED_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FACE_WITH_NO_GOOD_GESTURE
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F57A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F57A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F57A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F57A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F57A #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FLEXED_BICEPS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FLEXED_BICEPS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FLEXED_BICEPS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FLEXED_BICEPS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FLEXED_BICEPS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\COUPLE_WITH_HEART #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\COUPLE_WITH_HEART #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\COUPLE_WITH_HEART #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\COUPLE_WITH_HEART #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\COUPLE_WITH_HEART #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\KISS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\KISS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\KISS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\KISS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\KISS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\NAIL_POLISH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\NAIL_POLISH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\NAIL_POLISH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\NAIL_POLISH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\NAIL_POLISH #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\DANCER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\DANCER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\DANCER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\DANCER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\DANCER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BABY_ANGEL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BABY_ANGEL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BABY_ANGEL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BABY_ANGEL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BABY_ANGEL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PRINCESS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PRINCESS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PRINCESS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PRINCESS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PRINCESS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BABY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BABY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BABY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BABY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BABY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\OLDER_WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\OLDER_WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\OLDER_WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\OLDER_WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\OLDER_WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\OLDER_MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\OLDER_MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\OLDER_MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\OLDER_MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\OLDER_MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN_WITH_GUA_PI_MAO #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN_WITH_GUA_PI_MAO #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN_WITH_GUA_PI_MAO #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN_WITH_GUA_PI_MAO #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN_WITH_GUA_PI_MAO #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\TWO_WOMEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\TWO_WOMEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\TWO_WOMEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\TWO_WOMEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\TWO_WOMEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\TWO_MEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\TWO_MEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\TWO_MEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\TWO_MEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\TWO_MEN_HOLDING_HANDS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN_AND_WOMAN_HOLDING_HANDS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN_AND_WOMAN_HOLDING_HANDS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN_AND_WOMAN_HOLDING_HANDS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN_AND_WOMAN_HOLDING_HANDS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN_AND_WOMAN_HOLDING_HANDS
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\GIRL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\GIRL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\GIRL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\GIRL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\GIRL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\BOY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\BOY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\BOY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\BOY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\BOY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\OPEN_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\OPEN_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\OPEN_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\OPEN_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\OPEN_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\CLAPPING_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\CLAPPING_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\CLAPPING_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\CLAPPING_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\CLAPPING_HANDS_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\THUMBS_DOWN_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\THUMBS_DOWN_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\THUMBS_DOWN_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\THUMBS_DOWN_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\THUMBS_DOWN_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\THUMBS_UP_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\THUMBS_UP_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\THUMBS_UP_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\THUMBS_UP_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\THUMBS_UP_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\OK_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\OK_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\OK_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\OK_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\OK_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WAVING_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WAVING_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WAVING_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WAVING_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WAVING_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FISTED_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FISTED_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FISTED_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FISTED_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FISTED_HAND_SIGN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\NOSE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\NOSE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\NOSE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\NOSE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\NOSE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\EAR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\EAR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\EAR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\EAR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\EAR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\HORSE_RACING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\HORSE_RACING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\HORSE_RACING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\HORSE_RACING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\HORSE_RACING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\SNOWBOARDER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\SNOWBOARDER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\SNOWBOARDER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\SNOWBOARDER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\SNOWBOARDER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\FATHER_CHRISTMAS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\FATHER_CHRISTMAS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\FATHER_CHRISTMAS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\FATHER_CHRISTMAS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\FATHER_CHRISTMAS #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WRITING_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WRITING_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WRITING_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WRITING_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WRITING_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\VICTORY_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\VICTORY_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\VICTORY_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\VICTORY_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\VICTORY_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\RAISED_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\RAISED_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\RAISED_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\RAISED_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\RAISED_HAND #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\RAISED_FIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\RAISED_FIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\RAISED_FIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\RAISED_FIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\RAISED_FIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WHITE_UP_POINTING_INDEX #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WHITE_UP_POINTING_INDEX #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WHITE_UP_POINTING_INDEX #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WHITE_UP_POINTING_INDEX #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WHITE_UP_POINTING_INDEX #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WAVING_BLACK_FLAG #\TAG_LATIN_SMALL_LETTER_G
              #\TAG_LATIN_SMALL_LETTER_B #\TAG_LATIN_SMALL_LETTER_W
              #\TAG_LATIN_SMALL_LETTER_L #\TAG_LATIN_SMALL_LETTER_S
              #\CANCEL_TAG)
             (#\WAVING_BLACK_FLAG #\TAG_LATIN_SMALL_LETTER_G
              #\TAG_LATIN_SMALL_LETTER_B #\TAG_LATIN_SMALL_LETTER_S
              #\TAG_LATIN_SMALL_LETTER_C #\TAG_LATIN_SMALL_LETTER_T
              #\CANCEL_TAG)
             (#\WAVING_BLACK_FLAG #\TAG_LATIN_SMALL_LETTER_G
              #\TAG_LATIN_SMALL_LETTER_B #\TAG_LATIN_SMALL_LETTER_E
              #\TAG_LATIN_SMALL_LETTER_N #\TAG_LATIN_SMALL_LETTER_G
              #\CANCEL_TAG)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Z
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Z
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Z
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Y
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Y
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_X
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_W
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_W
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_V
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_U
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_J)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_T
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_X)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_J)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_B)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_S
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_R
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_R
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_R
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_R
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_R
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_Q
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_P
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_O
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_N
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_X)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Q)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_M
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_B)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_L
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_K
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_J
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_J
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_J
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_J
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Q)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_I
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_H
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Q)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_B)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_G
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_J)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_F
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_E
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_J)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_D
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_X)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_P)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_K)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_C
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Y)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_V)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Q)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_N)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_J)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_H)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_B)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_B
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_A)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Z)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_X)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_W)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_U)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_T)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_S)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_R)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_Q)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_O)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_M)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_L)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_I)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_G)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_F)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_E)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_D)
             (#\REGIONAL_INDICATOR_SYMBOL_LETTER_A
              #\REGIONAL_INDICATOR_SYMBOL_LETTER_C)
             (#\9 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\8 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\7 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\6 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\5 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\4 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\3 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\2 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\1 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\0 #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\* #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\# #\VARIATION_SELECTOR-16 #\COMBINING_ENCLOSING_KEYCAP)
             (#\PASSENGER_SHIP #\VARIATION_SELECTOR-16)
             (#\SATELLITE #\VARIATION_SELECTOR-16)
             (#\SMALL_AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MOTOR_BOAT #\VARIATION_SELECTOR-16)
             (#\RAILWAY_TRACK #\VARIATION_SELECTOR-16)
             (#\MOTORWAY #\VARIATION_SELECTOR-16)
             (#\OIL_DRUM #\VARIATION_SELECTOR-16)
             (#\SHIELD #\VARIATION_SELECTOR-16)
             (#\HAMMER_AND_WRENCH #\VARIATION_SELECTOR-16)
             (#\BED #\VARIATION_SELECTOR-16)
             (#\BELLHOP_BELL #\VARIATION_SELECTOR-16)
             (#\SHOPPING_BAGS #\VARIATION_SELECTOR-16)
             (#\COUCH_AND_LAMP #\VARIATION_SELECTOR-16)
             (#\WORLD_MAP #\VARIATION_SELECTOR-16)
             (#\BALLOT_BOX_WITH_BALLOT #\VARIATION_SELECTOR-16)
             (#\RIGHT_ANGER_BUBBLE #\VARIATION_SELECTOR-16)
             (#\LEFT_SPEECH_BUBBLE #\VARIATION_SELECTOR-16)
             (#\SPEAKING_HEAD_IN_SILHOUETTE #\VARIATION_SELECTOR-16)
             (#\DAGGER_KNIFE #\VARIATION_SELECTOR-16)
             (#\ROLLED-UP_NEWSPAPER #\VARIATION_SELECTOR-16)
             (#\OLD_KEY #\VARIATION_SELECTOR-16)
             (#\COMPRESSION #\VARIATION_SELECTOR-16)
             (#\SPIRAL_CALENDAR_PAD #\VARIATION_SELECTOR-16)
             (#\SPIRAL_NOTE_PAD #\VARIATION_SELECTOR-16)
             (#\WASTEBASKET #\VARIATION_SELECTOR-16)
             (#\FILE_CABINET #\VARIATION_SELECTOR-16)
             (#\CARD_FILE_BOX #\VARIATION_SELECTOR-16)
             (#\CARD_INDEX_DIVIDERS #\VARIATION_SELECTOR-16)
             (#\FRAME_WITH_PICTURE #\VARIATION_SELECTOR-16)
             (#\TRACKBALL #\VARIATION_SELECTOR-16)
             (#\THREE_BUTTON_MOUSE #\VARIATION_SELECTOR-16)
             (#\PRINTER #\VARIATION_SELECTOR-16)
             (#\DESKTOP_COMPUTER #\VARIATION_SELECTOR-16)
             (#\RAISED_HAND_WITH_FINGERS_SPLAYED #\VARIATION_SELECTOR-16)
             (#\LOWER_LEFT_CRAYON #\VARIATION_SELECTOR-16)
             (#\LOWER_LEFT_PAINTBRUSH #\VARIATION_SELECTOR-16)
             (#\LOWER_LEFT_FOUNTAIN_PEN #\VARIATION_SELECTOR-16)
             (#\LOWER_LEFT_BALLPOINT_PEN #\VARIATION_SELECTOR-16)
             (#\LINKED_PAPERCLIPS #\VARIATION_SELECTOR-16)
             (#\JOYSTICK #\VARIATION_SELECTOR-16)
             (#\SPIDER_WEB #\VARIATION_SELECTOR-16)
             (#\SPIDER #\VARIATION_SELECTOR-16)
             (#\DARK_SUNGLASSES #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\VARIATION_SELECTOR-16)
             (#\MAN_IN_BUSINESS_SUIT_LEVITATING #\VARIATION_SELECTOR-16)
             (#\HOLE #\VARIATION_SELECTOR-16)
             (#\MANTELPIECE_CLOCK #\VARIATION_SELECTOR-16)
             (#\CANDLE #\VARIATION_SELECTOR-16)
             (#\DOVE_OF_PEACE #\VARIATION_SELECTOR-16)
             (#\OM_SYMBOL #\VARIATION_SELECTOR-16)
             (#\FILM_PROJECTOR #\VARIATION_SELECTOR-16)
             (#\EYE #\VARIATION_SELECTOR-16)
             (#\CHIPMUNK #\VARIATION_SELECTOR-16)
             (#\LABEL #\VARIATION_SELECTOR-16)
             (#\ROSETTE #\VARIATION_SELECTOR-16)
             (#\WAVING_WHITE_FLAG #\VARIATION_SELECTOR-16)
             (#\STADIUM #\VARIATION_SELECTOR-16)
             (#\NATIONAL_PARK #\VARIATION_SELECTOR-16)
             (#\DESERT_ISLAND #\VARIATION_SELECTOR-16)
             (#\DESERT #\VARIATION_SELECTOR-16)
             (#\CLASSICAL_BUILDING #\VARIATION_SELECTOR-16)
             (#\DERELICT_HOUSE_BUILDING #\VARIATION_SELECTOR-16)
             (#\CITYSCAPE #\VARIATION_SELECTOR-16)
             (#\HOUSE_BUILDINGS #\VARIATION_SELECTOR-16)
             (#\BUILDING_CONSTRUCTION #\VARIATION_SELECTOR-16)
             (#\BEACH_WITH_UMBRELLA #\VARIATION_SELECTOR-16)
             (#\CAMPING #\VARIATION_SELECTOR-16)
             (#\SNOW_CAPPED_MOUNTAIN #\VARIATION_SELECTOR-16)
             (#\RACING_CAR #\VARIATION_SELECTOR-16)
             (#\RACING_MOTORCYCLE #\VARIATION_SELECTOR-16)
             (#\GOLFER #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\VARIATION_SELECTOR-16)
             (#\ADMISSION_TICKETS #\VARIATION_SELECTOR-16)
             (#\FILM_FRAMES #\VARIATION_SELECTOR-16)
             (#\CONTROL_KNOBS #\VARIATION_SELECTOR-16)
             (#\LEVEL_SLIDER #\VARIATION_SELECTOR-16)
             (#\STUDIO_MICROPHONE #\VARIATION_SELECTOR-16)
             (#\REMINDER_RIBBON #\VARIATION_SELECTOR-16)
             (#\MILITARY_MEDAL #\VARIATION_SELECTOR-16)
             (#\FORK_AND_KNIFE_WITH_PLATE #\VARIATION_SELECTOR-16)
             (#\HOT_PEPPER #\VARIATION_SELECTOR-16)
             (#\WIND_BLOWING_FACE #\VARIATION_SELECTOR-16)
             (#\FOG #\VARIATION_SELECTOR-16)
             (#\CLOUD_WITH_TORNADO #\VARIATION_SELECTOR-16)
             (#\CLOUD_WITH_LIGHTNING #\VARIATION_SELECTOR-16)
             (#\CLOUD_WITH_SNOW #\VARIATION_SELECTOR-16)
             (#\CLOUD_WITH_RAIN #\VARIATION_SELECTOR-16)
             (#\WHITE_SUN_BEHIND_CLOUD_WITH_RAIN #\VARIATION_SELECTOR-16)
             (#\WHITE_SUN_BEHIND_CLOUD #\VARIATION_SELECTOR-16)
             (#\WHITE_SUN_WITH_SMALL_CLOUD #\VARIATION_SELECTOR-16)
             (#\THERMOMETER #\VARIATION_SELECTOR-16)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-6708 #\VARIATION_SELECTOR-16)
             (#\SQUARED_KATAKANA_SA #\VARIATION_SELECTOR-16)
             (#\NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_P
              #\VARIATION_SELECTOR-16)
             (#\NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_O
              #\VARIATION_SELECTOR-16)
             (#\NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_B
              #\VARIATION_SELECTOR-16)
             (#\NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_A
              #\VARIATION_SELECTOR-16)
             (#\CIRCLED_IDEOGRAPH_SECRET #\VARIATION_SELECTOR-16)
             (#\CIRCLED_IDEOGRAPH_CONGRATULATION #\VARIATION_SELECTOR-16)
             (#\PART_ALTERNATION_MARK #\VARIATION_SELECTOR-16)
             (#\WAVY_DASH #\VARIATION_SELECTOR-16)
             (#\DOWNWARDS_BLACK_ARROW #\VARIATION_SELECTOR-16)
             (#\UPWARDS_BLACK_ARROW #\VARIATION_SELECTOR-16)
             (#\LEFTWARDS_BLACK_ARROW #\VARIATION_SELECTOR-16)
             (#\ARROW_POINTING_RIGHTWARDS_THEN_CURVING_DOWNWARDS
              #\VARIATION_SELECTOR-16)
             (#\ARROW_POINTING_RIGHTWARDS_THEN_CURVING_UPWARDS
              #\VARIATION_SELECTOR-16)
             (#\BLACK_RIGHTWARDS_ARROW #\VARIATION_SELECTOR-16)
             (#\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16)
             (#\HEAVY_HEART_EXCLAMATION_MARK_ORNAMENT #\VARIATION_SELECTOR-16)
             (#\SPARKLE #\VARIATION_SELECTOR-16)
             (#\SNOWFLAKE #\VARIATION_SELECTOR-16)
             (#\EIGHT_POINTED_BLACK_STAR #\VARIATION_SELECTOR-16)
             (#\EIGHT_SPOKED_ASTERISK #\VARIATION_SELECTOR-16)
             (#\STAR_OF_DAVID #\VARIATION_SELECTOR-16)
             (#\LATIN_CROSS #\VARIATION_SELECTOR-16)
             (#\HEAVY_MULTIPLICATION_X #\VARIATION_SELECTOR-16)
             (#\HEAVY_CHECK_MARK #\VARIATION_SELECTOR-16)
             (#\BLACK_NIB #\VARIATION_SELECTOR-16)
             (#\PENCIL #\VARIATION_SELECTOR-16)
             (#\WRITING_HAND #\VARIATION_SELECTOR-16)
             (#\VICTORY_HAND #\VARIATION_SELECTOR-16)
             (#\ENVELOPE #\VARIATION_SELECTOR-16)
             (#\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\BLACK_SCISSORS #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\VARIATION_SELECTOR-16)
             (#\ICE_SKATE #\VARIATION_SELECTOR-16)
             (#\SKIER #\VARIATION_SELECTOR-16)
             (#\FERRY #\VARIATION_SELECTOR-16)
             (#\UMBRELLA_ON_GROUND #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN #\VARIATION_SELECTOR-16)
             (#\SHINTO_SHRINE #\VARIATION_SELECTOR-16)
             (#\CHAINS #\VARIATION_SELECTOR-16)
             (#\HELMET_WITH_WHITE_CROSS #\VARIATION_SELECTOR-16)
             (#\PICK #\VARIATION_SELECTOR-16)
             (#\THUNDER_CLOUD_AND_RAIN #\VARIATION_SELECTOR-16)
             (#\FUNERAL_URN #\VARIATION_SELECTOR-16)
             (#\COFFIN #\VARIATION_SELECTOR-16)
             (#\MALE_WITH_STROKE_AND_MALE_AND_FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\WARNING_SIGN #\VARIATION_SELECTOR-16)
             (#\FLEUR-DE-LIS #\VARIATION_SELECTOR-16)
             (#\ATOM_SYMBOL #\VARIATION_SELECTOR-16)
             (#\GEAR #\VARIATION_SELECTOR-16)
             (#\ALEMBIC #\VARIATION_SELECTOR-16)
             (#\SCALES #\VARIATION_SELECTOR-16)
             (#\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\CROSSED_SWORDS #\VARIATION_SELECTOR-16)
             (#\HAMMER_AND_PICK #\VARIATION_SELECTOR-16)
             (#\PERMANENT_PAPER_SIGN #\VARIATION_SELECTOR-16)
             (#\BLACK_UNIVERSAL_RECYCLING_SYMBOL #\VARIATION_SELECTOR-16)
             (#\HOT_SPRINGS #\VARIATION_SELECTOR-16)
             (#\BLACK_DIAMOND_SUIT #\VARIATION_SELECTOR-16)
             (#\BLACK_HEART_SUIT #\VARIATION_SELECTOR-16)
             (#\BLACK_CLUB_SUIT #\VARIATION_SELECTOR-16)
             (#\BLACK_SPADE_SUIT #\VARIATION_SELECTOR-16)
             (#\BLACK_CHESS_PAWN #\VARIATION_SELECTOR-16)
             (#\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WHITE_SMILING_FACE #\VARIATION_SELECTOR-16)
             (#\WHITE_FROWNING_FACE #\VARIATION_SELECTOR-16)
             (#\WHEEL_OF_DHARMA #\VARIATION_SELECTOR-16)
             (#\YIN_YANG #\VARIATION_SELECTOR-16)
             (#\PEACE_SYMBOL #\VARIATION_SELECTOR-16)
             (#\STAR_AND_CRESCENT #\VARIATION_SELECTOR-16)
             (#\ORTHODOX_CROSS #\VARIATION_SELECTOR-16)
             (#\BIOHAZARD_SIGN #\VARIATION_SELECTOR-16)
             (#\RADIOACTIVE_SIGN #\VARIATION_SELECTOR-16)
             (#\SKULL_AND_CROSSBONES #\VARIATION_SELECTOR-16)
             (#\WHITE_UP_POINTING_INDEX #\VARIATION_SELECTOR-16)
             (#\SHAMROCK #\VARIATION_SELECTOR-16)
             (#\BALLOT_BOX_WITH_CHECK #\VARIATION_SELECTOR-16)
             (#\BLACK_TELEPHONE #\VARIATION_SELECTOR-16)
             (#\COMET #\VARIATION_SELECTOR-16)
             (#\SNOWMAN #\VARIATION_SELECTOR-16)
             (#\UMBRELLA #\VARIATION_SELECTOR-16)
             (#\CLOUD #\VARIATION_SELECTOR-16)
             (#\BLACK_SUN_WITH_RAYS #\VARIATION_SELECTOR-16)
             (#\BLACK_MEDIUM_SQUARE #\VARIATION_SELECTOR-16)
             (#\WHITE_MEDIUM_SQUARE #\VARIATION_SELECTOR-16)
             (#\BLACK_LEFT-POINTING_TRIANGLE #\VARIATION_SELECTOR-16)
             (#\BLACK_RIGHT-POINTING_TRIANGLE #\VARIATION_SELECTOR-16)
             (#\WHITE_SMALL_SQUARE #\VARIATION_SELECTOR-16)
             (#\BLACK_SMALL_SQUARE #\VARIATION_SELECTOR-16)
             (#\CIRCLED_LATIN_CAPITAL_LETTER_M #\VARIATION_SELECTOR-16)
             (#\BLACK_CIRCLE_FOR_RECORD #\VARIATION_SELECTOR-16)
             (#\BLACK_SQUARE_FOR_STOP #\VARIATION_SELECTOR-16)
             (#\DOUBLE_VERTICAL_BAR #\VARIATION_SELECTOR-16)
             (#\TIMER_CLOCK #\VARIATION_SELECTOR-16)
             (#\STOPWATCH #\VARIATION_SELECTOR-16)
             (#\BLACK_RIGHT-POINTING_TRIANGLE_WITH_DOUBLE_VERTICAL_BAR
              #\VARIATION_SELECTOR-16)
             (#\BLACK_LEFT-POINTING_DOUBLE_TRIANGLE_WITH_VERTICAL_BAR
              #\VARIATION_SELECTOR-16)
             (#\BLACK_RIGHT-POINTING_DOUBLE_TRIANGLE_WITH_VERTICAL_BAR
              #\VARIATION_SELECTOR-16)
             (#\EJECT_SYMBOL #\VARIATION_SELECTOR-16)
             (#\KEYBOARD #\VARIATION_SELECTOR-16)
             (#\RIGHTWARDS_ARROW_WITH_HOOK #\VARIATION_SELECTOR-16)
             (#\LEFTWARDS_ARROW_WITH_HOOK #\VARIATION_SELECTOR-16)
             (#\SOUTH_WEST_ARROW #\VARIATION_SELECTOR-16)
             (#\SOUTH_EAST_ARROW #\VARIATION_SELECTOR-16)
             (#\NORTH_EAST_ARROW #\VARIATION_SELECTOR-16)
             (#\NORTH_WEST_ARROW #\VARIATION_SELECTOR-16)
             (#\UP_DOWN_ARROW #\VARIATION_SELECTOR-16)
             (#\LEFT_RIGHT_ARROW #\VARIATION_SELECTOR-16)
             (#\INFORMATION_SOURCE #\VARIATION_SELECTOR-16)
             (#\TRADE_MARK_SIGN #\VARIATION_SELECTOR-16)
             (#\EXCLAMATION_QUESTION_MARK #\VARIATION_SELECTOR-16)
             (#\DOUBLE_EXCLAMATION_MARK #\VARIATION_SELECTOR-16)
             (#\REGISTERED_SIGN #\VARIATION_SELECTOR-16)
             (#\COPYRIGHT_SIGN #\VARIATION_SELECTOR-16) (#\U1FAD6) (#\U1FAD5)
             (#\U1FAD4) (#\U1FAD3) (#\U1FAD2) (#\U1FAD1) (#\U1FAD0) (#\U1FAC2)
             (#\U1FAC1) (#\U1FAC0) (#\U1FAB6) (#\U1FAB5) (#\U1FAB4) (#\U1FAB3)
             (#\U1FAB2) (#\U1FAB1) (#\U1FAB0) (#\U1FAA8) (#\U1FAA7) (#\U1FAA6)
             (#\U1FAA5) (#\U1FAA4) (#\U1FAA3) (#\U1FAA2) (#\U1FAA1) (#\U1FAA0)
             (#\U1FA9F) (#\U1FA9E) (#\U1FA9D) (#\U1FA9C) (#\U1FA9B) (#\U1FA9A)
             (#\U1FA99) (#\U1FA98) (#\U1FA97) (#\U1FA96) (#\U1FA95) (#\U1FA94)
             (#\U1FA93) (#\U1FA92) (#\U1FA91) (#\U1FA90) (#\U1FA86) (#\U1FA85)
             (#\U1FA84) (#\U1FA83) (#\U1FA82) (#\U1FA81) (#\U1FA80) (#\U1FA7A)
             (#\U1FA79) (#\U1FA78) (#\U1FA74) (#\U1FA73) (#\U1FA72) (#\U1FA71)
             (#\U1FA70) (#\U1F9FF) (#\U1F9FE) (#\U1F9FD) (#\U1F9FC) (#\U1F9FB)
             (#\U1F9FA) (#\U1F9F9) (#\U1F9F8) (#\U1F9F7) (#\U1F9F6) (#\U1F9F5)
             (#\U1F9F4) (#\U1F9F3) (#\U1F9F2) (#\U1F9F1) (#\U1F9F0) (#\U1F9EF)
             (#\U1F9EE) (#\U1F9ED) (#\U1F9EC) (#\U1F9EB) (#\U1F9EA) (#\U1F9E9)
             (#\U1F9E8) (#\U1F9E7) (#\U1F9E6) (#\U1F9E5) (#\U1F9E4) (#\U1F9E3)
             (#\U1F9E2) (#\U1F9E1) (#\U1F9E0) (#\U1F9DF) (#\U1F9DE) (#\U1F9DD)
             (#\U1F9DC) (#\U1F9DB) (#\U1F9DA) (#\U1F9D9) (#\U1F9D8) (#\U1F9D7)
             (#\U1F9D6) (#\U1F9D5) (#\U1F9D4) (#\U1F9D3) (#\U1F9D2) (#\U1F9D1)
             (#\U1F9D0) (#\U1F9CF) (#\U1F9CE) (#\U1F9CD) (#\U1F9CB) (#\U1F9CA)
             (#\U1F9C9) (#\U1F9C8) (#\U1F9C7) (#\U1F9C6) (#\U1F9C5) (#\U1F9C4)
             (#\U1F9C3) (#\U1F9C2) (#\U1F9C1) (#\CHEESE_WEDGE) (#\U1F9BF)
             (#\U1F9BE) (#\U1F9BD) (#\U1F9BC) (#\U1F9BB) (#\U1F9BA) (#\U1F9B9)
             (#\U1F9B8) (#\U1F9B7) (#\U1F9B6) (#\U1F9B5) (#\U1F9B4) (#\U1F9B3)
             (#\U1F9B2) (#\U1F9B1) (#\U1F9B0) (#\U1F9AF) (#\U1F9AE) (#\U1F9AD)
             (#\U1F9AC) (#\U1F9AB) (#\U1F9AA) (#\U1F9A9) (#\U1F9A8) (#\U1F9A7)
             (#\U1F9A6) (#\U1F9A5) (#\U1F9A4) (#\U1F9A3) (#\U1F9A2) (#\U1F9A1)
             (#\U1F9A0) (#\U1F99F) (#\U1F99E) (#\U1F99D) (#\U1F99C) (#\U1F99B)
             (#\U1F99A) (#\U1F999) (#\U1F998) (#\U1F997) (#\U1F996) (#\U1F995)
             (#\U1F994) (#\U1F993) (#\U1F992) (#\U1F991) (#\U1F990) (#\U1F98F)
             (#\U1F98E) (#\U1F98D) (#\U1F98C) (#\U1F98B) (#\U1F98A) (#\U1F989)
             (#\U1F988) (#\U1F987) (#\U1F986) (#\U1F985) (#\UNICORN_FACE)
             (#\TURKEY) (#\SCORPION) (#\LION_FACE) (#\CRAB) (#\U1F97F)
             (#\U1F97E) (#\U1F97D) (#\U1F97C) (#\U1F97B) (#\U1F97A) (#\U1F978)
             (#\U1F977) (#\U1F976) (#\U1F975) (#\U1F974) (#\U1F973) (#\U1F972)
             (#\U1F971) (#\U1F970) (#\U1F96F) (#\U1F96E) (#\U1F96D) (#\U1F96C)
             (#\U1F96B) (#\U1F96A) (#\U1F969) (#\U1F968) (#\U1F967) (#\U1F966)
             (#\U1F965) (#\U1F964) (#\U1F963) (#\U1F962) (#\U1F961) (#\U1F960)
             (#\U1F95F) (#\U1F95E) (#\U1F95D) (#\U1F95C) (#\U1F95B) (#\U1F95A)
             (#\U1F959) (#\U1F958) (#\U1F957) (#\U1F956) (#\U1F955) (#\U1F954)
             (#\U1F953) (#\U1F952) (#\U1F951) (#\U1F950) (#\U1F94F) (#\U1F94E)
             (#\U1F94D) (#\U1F94C) (#\U1F94B) (#\U1F94A) (#\U1F949) (#\U1F948)
             (#\U1F947) (#\U1F945) (#\U1F944) (#\U1F943) (#\U1F942) (#\U1F941)
             (#\U1F940) (#\U1F93F) (#\U1F93E) (#\U1F93D) (#\U1F93C) (#\U1F93A)
             (#\U1F939) (#\U1F938) (#\U1F937) (#\U1F936) (#\U1F935) (#\U1F934)
             (#\U1F933) (#\U1F932) (#\U1F931) (#\U1F930) (#\U1F92F) (#\U1F92E)
             (#\U1F92D) (#\U1F92C) (#\U1F92B) (#\U1F92A) (#\U1F929) (#\U1F928)
             (#\U1F927) (#\U1F926) (#\U1F925) (#\U1F924) (#\U1F923) (#\U1F922)
             (#\U1F921) (#\U1F920) (#\U1F91F) (#\U1F91E) (#\U1F91D) (#\U1F91C)
             (#\U1F91B) (#\U1F91A) (#\U1F919) (#\SIGN_OF_THE_HORNS)
             (#\HUGGING_FACE) (#\ROBOT_FACE) (#\FACE_WITH_HEAD-BANDAGE)
             (#\THINKING_FACE) (#\NERD_FACE) (#\FACE_WITH_THERMOMETER)
             (#\MONEY-MOUTH_FACE) (#\ZIPPER-MOUTH_FACE) (#\U1F90F) (#\U1F90E)
             (#\U1F90D) (#\U1F90C) (#\U1F7EB) (#\U1F7EA) (#\U1F7E9) (#\U1F7E8)
             (#\U1F7E7) (#\U1F7E6) (#\U1F7E5) (#\U1F7E4) (#\U1F7E3) (#\U1F7E2)
             (#\U1F7E1) (#\U1F7E0) (#\U1F6FC) (#\U1F6FB) (#\U1F6FA) (#\U1F6F9)
             (#\U1F6F8) (#\U1F6F7) (#\U1F6F6) (#\U1F6F5) (#\U1F6F4)
             (#\AIRPLANE_ARRIVING) (#\AIRPLANE_DEPARTURE) (#\U1F6D7) (#\U1F6D6)
             (#\U1F6D5) (#\U1F6D2) (#\U1F6D1) (#\PLACE_OF_WORSHIP)
             (#\SLEEPING_ACCOMMODATION) (#\LEFT_LUGGAGE) (#\BAGGAGE_CLAIM)
             (#\CUSTOMS) (#\PASSPORT_CONTROL) (#\BATHTUB) (#\BATH) (#\SHOWER)
             (#\WATER_CLOSET) (#\TOILET) (#\BABY_SYMBOL) (#\RESTROOM)
             (#\WOMENS_SYMBOL) (#\MENS_SYMBOL) (#\CHILDREN_CROSSING)
             (#\NO_PEDESTRIANS) (#\PEDESTRIAN) (#\MOUNTAIN_BICYCLIST)
             (#\BICYCLIST) (#\NO_BICYCLES) (#\BICYCLE)
             (#\NON-POTABLE_WATER_SYMBOL) (#\POTABLE_WATER_SYMBOL)
             (#\DO_NOT_LITTER_SYMBOL) (#\PUT_LITTER_IN_ITS_PLACE_SYMBOL)
             (#\NO_SMOKING_SYMBOL) (#\SMOKING_SYMBOL) (#\NO_ENTRY_SIGN)
             (#\DOOR) (#\TRIANGULAR_FLAG_ON_POST)
             (#\POLICE_CARS_REVOLVING_LIGHT) (#\CONSTRUCTION_SIGN)
             (#\VERTICAL_TRAFFIC_LIGHT) (#\HORIZONTAL_TRAFFIC_LIGHT)
             (#\SPEEDBOAT) (#\ROWBOAT) (#\SHIP) (#\AERIAL_TRAMWAY)
             (#\MOUNTAIN_CABLEWAY) (#\SUSPENSION_RAILWAY) (#\MOUNTAIN_RAILWAY)
             (#\MONORAIL) (#\TRACTOR) (#\ARTICULATED_LORRY) (#\DELIVERY_TRUCK)
             (#\RECREATIONAL_VEHICLE) (#\ONCOMING_AUTOMOBILE) (#\AUTOMOBILE)
             (#\ONCOMING_TAXI) (#\TAXI) (#\ONCOMING_POLICE_CAR) (#\POLICE_CAR)
             (#\FIRE_ENGINE) (#\AMBULANCE) (#\MINIBUS) (#\BUS_STOP)
             (#\TROLLEYBUS) (#\ONCOMING_BUS) (#\BUS) (#\TRAM_CAR) (#\TRAM)
             (#\STATION) (#\LIGHT_RAIL) (#\METRO) (#\TRAIN)
             (#\HIGH-SPEED_TRAIN_WITH_BULLET_NOSE) (#\HIGH-SPEED_TRAIN)
             (#\RAILWAY_CAR) (#\STEAM_LOCOMOTIVE) (#\HELICOPTER) (#\ROCKET)
             (#\PERSON_WITH_FOLDED_HANDS) (#\PERSON_WITH_POUTING_FACE)
             (#\PERSON_FROWNING) (#\PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION)
             (#\HAPPY_PERSON_RAISING_ONE_HAND) (#\SPEAK-NO-EVIL_MONKEY)
             (#\HEAR-NO-EVIL_MONKEY) (#\SEE-NO-EVIL_MONKEY)
             (#\PERSON_BOWING_DEEPLY) (#\FACE_WITH_OK_GESTURE)
             (#\FACE_WITH_NO_GOOD_GESTURE) (#\FACE_WITH_ROLLING_EYES)
             (#\UPSIDE-DOWN_FACE) (#\SLIGHTLY_SMILING_FACE)
             (#\SLIGHTLY_FROWNING_FACE) (#\WEARY_CAT_FACE) (#\CRYING_CAT_FACE)
             (#\POUTING_CAT_FACE) (#\KISSING_CAT_FACE_WITH_CLOSED_EYES)
             (#\CAT_FACE_WITH_WRY_SMILE)
             (#\SMILING_CAT_FACE_WITH_HEART-SHAPED_EYES)
             (#\SMILING_CAT_FACE_WITH_OPEN_MOUTH)
             (#\CAT_FACE_WITH_TEARS_OF_JOY)
             (#\GRINNING_CAT_FACE_WITH_SMILING_EYES) (#\FACE_WITH_MEDICAL_MASK)
             (#\FACE_WITHOUT_MOUTH) (#\DIZZY_FACE) (#\SLEEPING_FACE)
             (#\FLUSHED_FACE) (#\ASTONISHED_FACE) (#\FACE_SCREAMING_IN_FEAR)
             (#\FACE_WITH_OPEN_MOUTH_AND_COLD_SWEAT) (#\HUSHED_FACE)
             (#\FACE_WITH_OPEN_MOUTH) (#\LOUDLY_CRYING_FACE) (#\GRIMACING_FACE)
             (#\TIRED_FACE) (#\SLEEPY_FACE) (#\WEARY_FACE) (#\FEARFUL_FACE)
             (#\ANGUISHED_FACE) (#\FROWNING_FACE_WITH_OPEN_MOUTH)
             (#\DISAPPOINTED_BUT_RELIEVED_FACE) (#\FACE_WITH_LOOK_OF_TRIUMPH)
             (#\PERSEVERING_FACE) (#\CRYING_FACE) (#\POUTING_FACE)
             (#\ANGRY_FACE) (#\WORRIED_FACE) (#\DISAPPOINTED_FACE)
             (#\FACE_WITH_STUCK-OUT_TONGUE_AND_TIGHTLY-CLOSED_EYES)
             (#\FACE_WITH_STUCK-OUT_TONGUE_AND_WINKING_EYE)
             (#\FACE_WITH_STUCK-OUT_TONGUE) (#\KISSING_FACE_WITH_CLOSED_EYES)
             (#\KISSING_FACE_WITH_SMILING_EYES) (#\FACE_THROWING_A_KISS)
             (#\KISSING_FACE) (#\CONFOUNDED_FACE) (#\CONFUSED_FACE)
             (#\PENSIVE_FACE) (#\FACE_WITH_COLD_SWEAT) (#\UNAMUSED_FACE)
             (#\EXPRESSIONLESS_FACE) (#\NEUTRAL_FACE) (#\SMIRKING_FACE)
             (#\SMILING_FACE_WITH_SUNGLASSES)
             (#\SMILING_FACE_WITH_HEART-SHAPED_EYES) (#\RELIEVED_FACE)
             (#\FACE_SAVOURING_DELICIOUS_FOOD)
             (#\SMILING_FACE_WITH_SMILING_EYES) (#\WINKING_FACE)
             (#\SMILING_FACE_WITH_HORNS) (#\SMILING_FACE_WITH_HALO)
             (#\SMILING_FACE_WITH_OPEN_MOUTH_AND_TIGHTLY-CLOSED_EYES)
             (#\SMILING_FACE_WITH_OPEN_MOUTH_AND_COLD_SWEAT)
             (#\SMILING_FACE_WITH_OPEN_MOUTH_AND_SMILING_EYES)
             (#\SMILING_FACE_WITH_OPEN_MOUTH) (#\FACE_WITH_TEARS_OF_JOY)
             (#\GRINNING_FACE_WITH_SMILING_EYES) (#\GRINNING_FACE) (#\MOYAI)
             (#\SILHOUETTE_OF_JAPAN) (#\STATUE_OF_LIBERTY) (#\TOKYO_TOWER)
             (#\MOUNT_FUJI) (#\U1F5A4)
             (#\RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS)
             (#\REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED) (#\U1F57A)
             (#\CLOCK_FACE_TWELVE-THIRTY) (#\CLOCK_FACE_ELEVEN-THIRTY)
             (#\CLOCK_FACE_TEN-THIRTY) (#\CLOCK_FACE_NINE-THIRTY)
             (#\CLOCK_FACE_EIGHT-THIRTY) (#\CLOCK_FACE_SEVEN-THIRTY)
             (#\CLOCK_FACE_SIX-THIRTY) (#\CLOCK_FACE_FIVE-THIRTY)
             (#\CLOCK_FACE_FOUR-THIRTY) (#\CLOCK_FACE_THREE-THIRTY)
             (#\CLOCK_FACE_TWO-THIRTY) (#\CLOCK_FACE_ONE-THIRTY)
             (#\CLOCK_FACE_TWELVE_OCLOCK) (#\CLOCK_FACE_ELEVEN_OCLOCK)
             (#\CLOCK_FACE_TEN_OCLOCK) (#\CLOCK_FACE_NINE_OCLOCK)
             (#\CLOCK_FACE_EIGHT_OCLOCK) (#\CLOCK_FACE_SEVEN_OCLOCK)
             (#\CLOCK_FACE_SIX_OCLOCK) (#\CLOCK_FACE_FIVE_OCLOCK)
             (#\CLOCK_FACE_FOUR_OCLOCK) (#\CLOCK_FACE_THREE_OCLOCK)
             (#\CLOCK_FACE_TWO_OCLOCK) (#\CLOCK_FACE_ONE_OCLOCK)
             (#\MENORAH_WITH_NINE_BRANCHES) (#\SYNAGOGUE) (#\MOSQUE) (#\KAABA)
             (#\DOWN-POINTING_SMALL_RED_TRIANGLE)
             (#\UP-POINTING_SMALL_RED_TRIANGLE) (#\DOWN-POINTING_RED_TRIANGLE)
             (#\UP-POINTING_RED_TRIANGLE) (#\SMALL_BLUE_DIAMOND)
             (#\SMALL_ORANGE_DIAMOND) (#\LARGE_BLUE_DIAMOND)
             (#\LARGE_ORANGE_DIAMOND) (#\LARGE_BLUE_CIRCLE)
             (#\LARGE_RED_CIRCLE) (#\WHITE_SQUARE_BUTTON)
             (#\BLACK_SQUARE_BUTTON) (#\TRIDENT_EMBLEM)
             (#\JAPANESE_SYMBOL_FOR_BEGINNER)
             (#\SIX_POINTED_STAR_WITH_MIDDLE_DOT) (#\CRYSTAL_BALL)
             (#\TELESCOPE) (#\MICROSCOPE) (#\PISTOL) (#\HOCHO) (#\NUT_AND_BOLT)
             (#\HAMMER) (#\WRENCH) (#\ELECTRIC_TORCH) (#\FIRE)
             (#\INPUT_SYMBOL_FOR_LATIN_LETTERS) (#\INPUT_SYMBOL_FOR_SYMBOLS)
             (#\INPUT_SYMBOL_FOR_NUMBERS)
             (#\INPUT_SYMBOL_FOR_LATIN_SMALL_LETTERS)
             (#\INPUT_SYMBOL_FOR_LATIN_CAPITAL_LETTERS) (#\KEYCAP_TEN)
             (#\NO_ONE_UNDER_EIGHTEEN_SYMBOL) (#\TOP_WITH_UPWARDS_ARROW_ABOVE)
             (#\SOON_WITH_RIGHTWARDS_ARROW_ABOVE)
             (#\ON_WITH_EXCLAMATION_MARK_WITH_LEFT_RIGHT_ARROW_ABOVE)
             (#\END_WITH_LEFTWARDS_ARROW_ABOVE)
             (#\BACK_WITH_LEFTWARDS_ARROW_ABOVE) (#\RADIO_BUTTON)
             (#\LINK_SYMBOL) (#\BOOKMARK) (#\BELL_WITH_CANCELLATION_STROKE)
             (#\BELL) (#\OPEN_LOCK) (#\LOCK) (#\KEY) (#\CLOSED_LOCK_WITH_KEY)
             (#\LOCK_WITH_INK_PEN) (#\RIGHT-POINTING_MAGNIFYING_GLASS)
             (#\LEFT-POINTING_MAGNIFYING_GLASS) (#\ELECTRIC_PLUG) (#\BATTERY)
             (#\SPEAKER_WITH_THREE_SOUND_WAVES) (#\SPEAKER_WITH_ONE_SOUND_WAVE)
             (#\SPEAKER) (#\SPEAKER_WITH_CANCELLATION_STROKE)
             (#\HIGH_BRIGHTNESS_SYMBOL) (#\LOW_BRIGHTNESS_SYMBOL)
             (#\ANTICLOCKWISE_DOWNWARDS_AND_UPWARDS_OPEN_CIRCLE_ARROWS)
             (#\CLOCKWISE_DOWNWARDS_AND_UPWARDS_OPEN_CIRCLE_ARROWS)
             (#\CLOCKWISE_RIGHTWARDS_AND_LEFTWARDS_OPEN_CIRCLE_ARROWS_WITH_CIRCLED_ONE_OVERLAY)
             (#\CLOCKWISE_RIGHTWARDS_AND_LEFTWARDS_OPEN_CIRCLE_ARROWS)
             (#\TWISTED_RIGHTWARDS_ARROWS) (#\PRAYER_BEADS) (#\VIDEOCASSETTE)
             (#\RADIO) (#\TELEVISION) (#\VIDEO_CAMERA) (#\CAMERA_WITH_FLASH)
             (#\CAMERA) (#\ANTENNA_WITH_BARS) (#\NO_MOBILE_PHONES)
             (#\MOBILE_PHONE_OFF) (#\VIBRATION_MODE)
             (#\MOBILE_PHONE_WITH_RIGHTWARDS_ARROW_AT_LEFT) (#\MOBILE_PHONE)
             (#\NEWSPAPER) (#\POSTAL_HORN) (#\POSTBOX)
             (#\OPEN_MAILBOX_WITH_LOWERED_FLAG)
             (#\OPEN_MAILBOX_WITH_RAISED_FLAG)
             (#\CLOSED_MAILBOX_WITH_RAISED_FLAG)
             (#\CLOSED_MAILBOX_WITH_LOWERED_FLAG)
             (#\ENVELOPE_WITH_DOWNWARDS_ARROW_ABOVE) (#\INCOMING_ENVELOPE)
             (#\E-MAIL_SYMBOL) (#\PACKAGE) (#\INBOX_TRAY) (#\OUTBOX_TRAY)
             (#\CHEERING_MEGAPHONE) (#\PUBLIC_ADDRESS_LOUDSPEAKER)
             (#\SATELLITE_ANTENNA) (#\FAX_MACHINE) (#\PAGER)
             (#\TELEPHONE_RECEIVER) (#\MEMO) (#\SCROLL) (#\NAME_BADGE)
             (#\BOOKS) (#\ORANGE_BOOK) (#\BLUE_BOOK) (#\GREEN_BOOK)
             (#\OPEN_BOOK) (#\CLOSED_BOOK) (#\NOTEBOOK_WITH_DECORATIVE_COVER)
             (#\NOTEBOOK) (#\LEDGER) (#\BOOKMARK_TABS) (#\TRIANGULAR_RULER)
             (#\STRAIGHT_RULER) (#\PAPERCLIP) (#\ROUND_PUSHPIN) (#\PUSHPIN)
             (#\CLIPBOARD) (#\BAR_CHART) (#\CHART_WITH_DOWNWARDS_TREND)
             (#\CHART_WITH_UPWARDS_TREND) (#\CARD_INDEX) (#\TEAR-OFF_CALENDAR)
             (#\CALENDAR) (#\PAGE_FACING_UP) (#\PAGE_WITH_CURL)
             (#\OPEN_FILE_FOLDER) (#\FILE_FOLDER) (#\DVD) (#\OPTICAL_DISC)
             (#\FLOPPY_DISK) (#\MINIDISC) (#\BRIEFCASE) (#\PERSONAL_COMPUTER)
             (#\SEAT) (#\CHART_WITH_UPWARDS_TREND_AND_YEN_SIGN)
             (#\MONEY_WITH_WINGS) (#\BANKNOTE_WITH_POUND_SIGN)
             (#\BANKNOTE_WITH_EURO_SIGN) (#\BANKNOTE_WITH_DOLLAR_SIGN)
             (#\BANKNOTE_WITH_YEN_SIGN) (#\CREDIT_CARD) (#\HEAVY_DOLLAR_SIGN)
             (#\CURRENCY_EXCHANGE) (#\MONEY_BAG) (#\HUNDRED_POINTS_SYMBOL)
             (#\WHITE_FLOWER) (#\THOUGHT_BALLOON) (#\SPEECH_BALLOON)
             (#\DIZZY_SYMBOL) (#\FLEXED_BICEPS) (#\PILE_OF_POO) (#\DASH_SYMBOL)
             (#\DROPLET) (#\SPLASHING_SWEAT_SYMBOL) (#\COLLISION_SYMBOL)
             (#\SLEEPING_SYMBOL) (#\BOMB) (#\ANGER_SYMBOL)
             (#\ELECTRIC_LIGHT_BULB) (#\DIAMOND_SHAPE_WITH_A_DOT_INSIDE)
             (#\HEART_DECORATION) (#\REVOLVING_HEARTS) (#\HEART_WITH_RIBBON)
             (#\PURPLE_HEART) (#\YELLOW_HEART) (#\GREEN_HEART) (#\BLUE_HEART)
             (#\HEART_WITH_ARROW) (#\GROWING_HEART) (#\SPARKLING_HEART)
             (#\TWO_HEARTS) (#\BROKEN_HEART) (#\BEATING_HEART) (#\WEDDING)
             (#\COUPLE_WITH_HEART) (#\BOUQUET) (#\KISS) (#\GEM_STONE) (#\RING)
             (#\LOVE_LETTER) (#\KISS_MARK) (#\PILL) (#\SYRINGE) (#\BARBER_POLE)
             (#\HAIRCUT) (#\FACE_MASSAGE) (#\NAIL_POLISH) (#\LIPSTICK)
             (#\DANCER) (#\GUARDSMAN) (#\INFORMATION_DESK_PERSON) (#\SKULL)
             (#\IMP) (#\ALIEN_MONSTER) (#\EXTRATERRESTRIAL_ALIEN)
             (#\BABY_ANGEL) (#\GHOST) (#\JAPANESE_GOBLIN) (#\JAPANESE_OGRE)
             (#\PRINCESS) (#\CONSTRUCTION_WORKER) (#\BABY) (#\OLDER_WOMAN)
             (#\OLDER_MAN) (#\MAN_WITH_TURBAN) (#\MAN_WITH_GUA_PI_MAO)
             (#\PERSON_WITH_BLOND_HAIR) (#\BRIDE_WITH_VEIL)
             (#\WOMAN_WITH_BUNNY_EARS) (#\POLICE_OFFICER)
             (#\TWO_WOMEN_HOLDING_HANDS) (#\TWO_MEN_HOLDING_HANDS)
             (#\MAN_AND_WOMAN_HOLDING_HANDS) (#\FAMILY) (#\WOMAN) (#\MAN)
             (#\GIRL) (#\BOY) (#\BUSTS_IN_SILHOUETTE) (#\BUST_IN_SILHOUETTE)
             (#\FOOTPRINTS) (#\WOMANS_BOOTS) (#\WOMANS_SANDAL)
             (#\HIGH-HEELED_SHOE) (#\ATHLETIC_SHOE) (#\MANS_SHOE) (#\POUCH)
             (#\HANDBAG) (#\PURSE) (#\WOMANS_CLOTHES) (#\BIKINI) (#\KIMONO)
             (#\DRESS) (#\JEANS) (#\T-SHIRT) (#\NECKTIE) (#\EYEGLASSES)
             (#\WOMANS_HAT) (#\CROWN) (#\OPEN_HANDS_SIGN)
             (#\CLAPPING_HANDS_SIGN) (#\THUMBS_DOWN_SIGN) (#\THUMBS_UP_SIGN)
             (#\OK_HAND_SIGN) (#\WAVING_HAND_SIGN) (#\FISTED_HAND_SIGN)
             (#\WHITE_RIGHT_POINTING_BACKHAND_INDEX)
             (#\WHITE_LEFT_POINTING_BACKHAND_INDEX)
             (#\WHITE_DOWN_POINTING_BACKHAND_INDEX)
             (#\WHITE_UP_POINTING_BACKHAND_INDEX) (#\TONGUE) (#\MOUTH) (#\NOSE)
             (#\EAR) (#\EYES) (#\PAW_PRINTS) (#\PIG_NOSE) (#\PANDA_FACE)
             (#\BEAR_FACE) (#\WOLF_FACE) (#\HAMSTER_FACE) (#\FROG_FACE)
             (#\PIG_FACE) (#\DOG_FACE) (#\MONKEY_FACE) (#\HORSE_FACE)
             (#\SPOUTING_WHALE) (#\DRAGON_FACE) (#\CAT_FACE) (#\RABBIT_FACE)
             (#\TIGER_FACE) (#\COW_FACE) (#\MOUSE_FACE) (#\DOLPHIN)
             (#\BACTRIAN_CAMEL) (#\DROMEDARY_CAMEL) (#\POODLE) (#\KOALA)
             (#\PENGUIN) (#\BIRD) (#\FRONT-FACING_BABY_CHICK) (#\BABY_CHICK)
             (#\HATCHING_CHICK) (#\TURTLE) (#\BLOWFISH) (#\TROPICAL_FISH)
             (#\FISH) (#\LADY_BEETLE) (#\HONEYBEE) (#\ANT) (#\BUG)
             (#\SPIRAL_SHELL) (#\OCTOPUS) (#\ELEPHANT) (#\BOAR) (#\PIG) (#\DOG)
             (#\CHICKEN) (#\ROOSTER) (#\MONKEY) (#\SHEEP) (#\GOAT) (#\RAM)
             (#\HORSE) (#\SNAKE) (#\SNAIL) (#\WHALE) (#\CROCODILE) (#\DRAGON)
             (#\CAT) (#\RABBIT) (#\LEOPARD) (#\TIGER) (#\COW) (#\WATER_BUFFALO)
             (#\OX) (#\MOUSE) (#\RAT) (#\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2) (#\AMPHORA)
             (#\BOW_AND_ARROW) (#\BADMINTON_RACQUET_AND_SHUTTLECOCK)
             (#\WAVING_BLACK_FLAG) (#\EUROPEAN_CASTLE) (#\JAPANESE_CASTLE)
             (#\IZAKAYA_LANTERN) (#\FACTORY) (#\DEPARTMENT_STORE) (#\SCHOOL)
             (#\CONVENIENCE_STORE) (#\LOVE_HOTEL) (#\HOTEL)
             (#\AUTOMATED_TELLER_MACHINE) (#\BANK) (#\HOSPITAL)
             (#\EUROPEAN_POST_OFFICE) (#\JAPANESE_POST_OFFICE)
             (#\OFFICE_BUILDING) (#\HOUSE_WITH_GARDEN) (#\HOUSE_BUILDING)
             (#\TABLE_TENNIS_PADDLE_AND_BALL) (#\ICE_HOCKEY_STICK_AND_PUCK)
             (#\FIELD_HOCKEY_STICK_AND_BALL) (#\VOLLEYBALL)
             (#\CRICKET_BAT_AND_BALL) (#\SWIMMER) (#\RUGBY_FOOTBALL)
             (#\AMERICAN_FOOTBALL) (#\HORSE_RACING) (#\TROPHY) (#\SPORTS_MEDAL)
             (#\SURFER) (#\RUNNER) (#\SNOWBOARDER) (#\CHEQUERED_FLAG)
             (#\BASKETBALL_AND_HOOP) (#\SKI_AND_SKI_BOOT)
             (#\TENNIS_RACQUET_AND_BALL) (#\RUNNING_SHIRT_WITH_SASH)
             (#\MUSICAL_SCORE) (#\VIOLIN) (#\TRUMPET) (#\MUSICAL_KEYBOARD)
             (#\GUITAR) (#\SAXOPHONE) (#\MULTIPLE_MUSICAL_NOTES)
             (#\MUSICAL_NOTE) (#\FLOWER_PLAYING_CARDS) (#\BOWLING) (#\GAME_DIE)
             (#\BILLIARDS) (#\SLOT_MACHINE) (#\DIRECT_HIT) (#\VIDEO_GAME)
             (#\PERFORMING_ARTS) (#\CLAPPER_BOARD) (#\TICKET) (#\CIRCUS_TENT)
             (#\TOP_HAT) (#\ARTIST_PALETTE) (#\HEADPHONE) (#\CINEMA)
             (#\MOVIE_CAMERA) (#\MICROPHONE) (#\FISHING_POLE_AND_FISH)
             (#\ROLLER_COASTER) (#\FERRIS_WHEEL) (#\CAROUSEL_HORSE)
             (#\GRADUATION_CAP) (#\SCHOOL_SATCHEL) (#\MOON_VIEWING_CEREMONY)
             (#\WIND_CHIME) (#\CARP_STREAMER) (#\JAPANESE_DOLLS)
             (#\PINE_DECORATION) (#\CROSSED_FLAGS) (#\TANABATA_TREE)
             (#\CONFETTI_BALL) (#\PARTY_POPPER) (#\BALLOON)
             (#\FIREWORK_SPARKLER) (#\FIREWORKS) (#\FATHER_CHRISTMAS)
             (#\CHRISTMAS_TREE) (#\JACK-O-LANTERN) (#\BIRTHDAY_CAKE)
             (#\WRAPPED_PRESENT) (#\RIBBON) (#\POPCORN)
             (#\BOTTLE_WITH_POPPING_CORK) (#\BABY_BOTTLE)
             (#\CLINKING_BEER_MUGS) (#\BEER_MUG) (#\TROPICAL_DRINK)
             (#\COCKTAIL_GLASS) (#\WINE_GLASS) (#\SAKE_BOTTLE_AND_CUP)
             (#\TEACUP_WITHOUT_HANDLE) (#\FORK_AND_KNIFE) (#\COOKING)
             (#\POT_OF_FOOD) (#\BENTO_BOX) (#\SHORTCAKE) (#\HONEY_POT)
             (#\CUSTARD) (#\LOLLIPOP) (#\CANDY) (#\CHOCOLATE_BAR) (#\COOKIE)
             (#\DOUGHNUT) (#\ICE_CREAM) (#\SHAVED_ICE) (#\SOFT_ICE_CREAM)
             (#\FISH_CAKE_WITH_SWIRL_DESIGN) (#\FRIED_SHRIMP) (#\SUSHI)
             (#\ODEN) (#\DANGO) (#\ROASTED_SWEET_POTATO) (#\FRENCH_FRIES)
             (#\BREAD) (#\SPAGHETTI) (#\STEAMING_BOWL) (#\CURRY_AND_RICE)
             (#\COOKED_RICE) (#\RICE_BALL) (#\RICE_CRACKER) (#\POULTRY_LEG)
             (#\MEAT_ON_BONE) (#\SLICE_OF_PIZZA) (#\HAMBURGER) (#\STRAWBERRY)
             (#\CHERRIES) (#\PEACH) (#\PEAR) (#\GREEN_APPLE) (#\RED_APPLE)
             (#\PINEAPPLE) (#\BANANA) (#\LEMON) (#\TANGERINE) (#\WATERMELON)
             (#\MELON) (#\GRAPES) (#\AUBERGINE) (#\TOMATO) (#\MUSHROOM)
             (#\LEAF_FLUTTERING_IN_WIND) (#\FALLEN_LEAF) (#\MAPLE_LEAF)
             (#\FOUR_LEAF_CLOVER) (#\HERB) (#\EAR_OF_RICE) (#\EAR_OF_MAIZE)
             (#\BLOSSOM) (#\SUNFLOWER) (#\HIBISCUS) (#\ROSE) (#\CHERRY_BLOSSOM)
             (#\TULIP) (#\CACTUS) (#\PALM_TREE) (#\DECIDUOUS_TREE)
             (#\EVERGREEN_TREE) (#\SEEDLING) (#\CHESTNUT) (#\BURRITO) (#\TACO)
             (#\HOT_DOG) (#\SHOOTING_STAR) (#\GLOWING_STAR) (#\SUN_WITH_FACE)
             (#\FULL_MOON_WITH_FACE) (#\LAST_QUARTER_MOON_WITH_FACE)
             (#\FIRST_QUARTER_MOON_WITH_FACE) (#\NEW_MOON_WITH_FACE)
             (#\CRESCENT_MOON) (#\WANING_CRESCENT_MOON_SYMBOL)
             (#\LAST_QUARTER_MOON_SYMBOL) (#\WANING_GIBBOUS_MOON_SYMBOL)
             (#\FULL_MOON_SYMBOL) (#\WAXING_GIBBOUS_MOON_SYMBOL)
             (#\FIRST_QUARTER_MOON_SYMBOL) (#\WAXING_CRESCENT_MOON_SYMBOL)
             (#\NEW_MOON_SYMBOL) (#\GLOBE_WITH_MERIDIANS)
             (#\EARTH_GLOBE_ASIA-AUSTRALIA) (#\EARTH_GLOBE_AMERICAS)
             (#\EARTH_GLOBE_EUROPE-AFRICA) (#\MILKY_WAY) (#\VOLCANO)
             (#\WATER_WAVE) (#\BRIDGE_AT_NIGHT) (#\RAINBOW)
             (#\SUNSET_OVER_BUILDINGS) (#\CITYSCAPE_AT_DUSK) (#\SUNRISE)
             (#\SUNRISE_OVER_MOUNTAINS) (#\NIGHT_WITH_STARS)
             (#\CLOSED_UMBRELLA) (#\FOGGY) (#\CYCLONE)
             (#\CIRCLED_IDEOGRAPH_ACCEPT) (#\CIRCLED_IDEOGRAPH_ADVANTAGE)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-55B6)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-5272)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-7533)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-6709)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-6E80)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-5408)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-7A7A)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-7981)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-6307)
             (#\SQUARED_CJK_UNIFIED_IDEOGRAPH-7121) (#\SQUARED_KATAKANA_KOKO)
             (#\SQUARED_VS) (#\SQUARED_UP_WITH_EXCLAMATION_MARK)
             (#\SQUARED_SOS) (#\SQUARED_OK) (#\SQUARED_NG) (#\SQUARED_NEW)
             (#\SQUARED_ID) (#\SQUARED_FREE) (#\SQUARED_COOL) (#\SQUARED_CL)
             (#\NEGATIVE_SQUARED_AB) (#\PLAYING_CARD_BLACK_JOKER)
             (#\MAHJONG_TILE_RED_DRAGON) (#\HEAVY_LARGE_CIRCLE)
             (#\WHITE_MEDIUM_STAR) (#\WHITE_LARGE_SQUARE)
             (#\BLACK_LARGE_SQUARE) (#\DOUBLE_CURLY_LOOP) (#\CURLY_LOOP)
             (#\HEAVY_DIVISION_SIGN) (#\HEAVY_MINUS_SIGN) (#\HEAVY_PLUS_SIGN)
             (#\HEAVY_EXCLAMATION_MARK_SYMBOL)
             (#\WHITE_EXCLAMATION_MARK_ORNAMENT)
             (#\WHITE_QUESTION_MARK_ORNAMENT) (#\BLACK_QUESTION_MARK_ORNAMENT)
             (#\NEGATIVE_SQUARED_CROSS_MARK) (#\CROSS_MARK) (#\SPARKLES)
             (#\RAISED_HAND) (#\RAISED_FIST) (#\WHITE_HEAVY_CHECK_MARK)
             (#\FUEL_PUMP) (#\TENT) (#\SAILBOAT) (#\FLAG_IN_HOLE) (#\FOUNTAIN)
             (#\CHURCH) (#\NO_ENTRY) (#\OPHIUCHUS) (#\SUN_BEHIND_CLOUD)
             (#\SNOWMAN_WITHOUT_SNOW) (#\BASEBALL) (#\SOCCER_BALL)
             (#\MEDIUM_BLACK_CIRCLE) (#\MEDIUM_WHITE_CIRCLE)
             (#\HIGH_VOLTAGE_SIGN) (#\ANCHOR) (#\WHEELCHAIR_SYMBOL) (#\PISCES)
             (#\AQUARIUS) (#\CAPRICORN) (#\SAGITTARIUS) (#\SCORPIUS) (#\LIBRA)
             (#\VIRGO) (#\LEO) (#\CANCER) (#\GEMINI) (#\TAURUS) (#\ARIES)
             (#\HOT_BEVERAGE) (#\UMBRELLA_WITH_RAIN_DROPS)
             (#\BLACK_MEDIUM_SMALL_SQUARE) (#\WHITE_MEDIUM_SMALL_SQUARE)
             (#\HOURGLASS_WITH_FLOWING_SAND) (#\ALARM_CLOCK)
             (#\BLACK_DOWN-POINTING_DOUBLE_TRIANGLE)
             (#\BLACK_UP-POINTING_DOUBLE_TRIANGLE)
             (#\BLACK_LEFT-POINTING_DOUBLE_TRIANGLE)
             (#\BLACK_RIGHT-POINTING_DOUBLE_TRIANGLE) (#\HOURGLASS) (#\WATCH))))

;;;; expansion of (gen-matcher-zwj-sequences)

(defun emoji-zwj-sequences-p (chars)
  (member-if (lambda (a) (match-codepoint-p chars a))
           '((#\U1F9D1 #\ZERO_WIDTH_JOINER #\CHRISTMAS_TREE)
             (#\FACE_WITHOUT_MOUTH #\ZERO_WIDTH_JOINER #\FOG
              #\VARIATION_SELECTOR-16)
             (#\DIZZY_FACE #\ZERO_WIDTH_JOINER #\DIZZY_SYMBOL)
             (#\FACE_WITH_OPEN_MOUTH #\ZERO_WIDTH_JOINER #\DASH_SYMBOL)
             (#\EYE #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\LEFT_SPEECH_BUBBLE #\VARIATION_SELECTOR-16)
             (#\BEAR_FACE #\ZERO_WIDTH_JOINER #\SNOWFLAKE
              #\VARIATION_SELECTOR-16)
             (#\DOG #\ZERO_WIDTH_JOINER #\U1F9BA)
             (#\CAT #\ZERO_WIDTH_JOINER #\BLACK_LARGE_SQUARE)
             (#\WAVING_BLACK_FLAG #\ZERO_WIDTH_JOINER #\SKULL_AND_CROSSBONES
              #\VARIATION_SELECTOR-16)
             (#\WAVING_WHITE_FLAG #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\RAINBOW)
             (#\WAVING_WHITE_FLAG #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MALE_WITH_STROKE_AND_MALE_AND_FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1FA79)
             (#\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\FIRE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9B3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9B2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9B1)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9B0)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9B3)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9B2)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9B1)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9B0)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9B3)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9B2)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9B1)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9B0)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B1)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9B0)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9B3)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9B2)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9B1)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9B0)
             (#\U1F9DF #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DF #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9DE #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DD #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DC #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DB #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9DA #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D9 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D8 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D7 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D6 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D4 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CF #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9CD #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B9 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9B8 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93E #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93D #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F93C #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F93C #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F939 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F938 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F937 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F935 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F926 #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PEDESTRIAN #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\MOUNTAIN_BICYCLIST #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\BICYCLIST #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\ROWBOAT #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_POUTING_FACE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_FROWNING #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\HAPPY_PERSON_RAISING_ONE_HAND #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_BOWING_DEEPLY #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_OK_GESTURE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_WITH_NO_GOOD_GESTURE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SLEUTH_OR_SPY #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\HAIRCUT #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\FACE_MASSAGE #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\GUARDSMAN #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\INFORMATION_DESK_PERSON #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\CONSTRUCTION_WORKER #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\MAN_WITH_TURBAN #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BLOND_HAIR #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\BRIDE_WITH_VEIL #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\WOMAN_WITH_BUNNY_EARS #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\WOMAN_WITH_BUNNY_EARS #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\POLICE_OFFICER #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\GOLFER #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\GOLFER #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\GOLFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\WEIGHT_LIFTER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\ZERO_WIDTH_JOINER #\MALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\SWIMMER #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\SURFER #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\RUNNER #\ZERO_WIDTH_JOINER #\FEMALE_SIGN
              #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MALE_SIGN #\VARIATION_SELECTOR-16)
             (#\PERSON_WITH_BALL #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FEMALE_SIGN #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9BD)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9BC)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F9AF)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FIRE_ENGINE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\ROCKET)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MICROSCOPE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\WRENCH)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\BRIEFCASE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\FACTORY)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\SCHOOL)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\ARTIST_PALETTE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\MICROPHONE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\GRADUATION_CAP)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\BABY_BOTTLE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\COOKING)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\EAR_OF_RICE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\STAFF_OF_AESCULAPIUS
              #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9BD)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9BC)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F9AF)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\FIRE_ENGINE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\ROCKET)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\MICROSCOPE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\WRENCH)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\BRIEFCASE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\PERSONAL_COMPUTER)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\FACTORY)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\SCHOOL)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\ARTIST_PALETTE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\MICROPHONE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\GRADUATION_CAP)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\BABY_BOTTLE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\COOKING)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\EAR_OF_RICE)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\SCALES #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\STAFF_OF_AESCULAPIUS
              #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9BD)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9BC)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\U1F9AF)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\FIRE_ENGINE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\ROCKET)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\MICROSCOPE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WRENCH)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\BRIEFCASE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\PERSONAL_COMPUTER)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\FACTORY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\SCHOOL)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\ARTIST_PALETTE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\MICROPHONE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\GRADUATION_CAP)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\BABY_BOTTLE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\COOKING)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\EAR_OF_RICE)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\SCALES #\VARIATION_SELECTOR-16)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\STAFF_OF_AESCULAPIUS
              #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9BD)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9BC)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F9AF)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\FIRE_ENGINE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\ROCKET)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\MICROSCOPE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\WRENCH)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\BRIEFCASE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\PERSONAL_COMPUTER)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\FACTORY)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\SCHOOL)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\ARTIST_PALETTE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\MICROPHONE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\GRADUATION_CAP)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\BABY_BOTTLE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\COOKING)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\EAR_OF_RICE)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\STAFF_OF_AESCULAPIUS #\VARIATION_SELECTOR-16)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9BD)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9BC)
             (#\MAN #\ZERO_WIDTH_JOINER #\U1F9AF)
             (#\MAN #\ZERO_WIDTH_JOINER #\FIRE_ENGINE)
             (#\MAN #\ZERO_WIDTH_JOINER #\ROCKET)
             (#\MAN #\ZERO_WIDTH_JOINER #\MICROSCOPE)
             (#\MAN #\ZERO_WIDTH_JOINER #\WRENCH)
             (#\MAN #\ZERO_WIDTH_JOINER #\BRIEFCASE)
             (#\MAN #\ZERO_WIDTH_JOINER #\PERSONAL_COMPUTER)
             (#\MAN #\ZERO_WIDTH_JOINER #\FACTORY)
             (#\MAN #\ZERO_WIDTH_JOINER #\SCHOOL)
             (#\MAN #\ZERO_WIDTH_JOINER #\ARTIST_PALETTE)
             (#\MAN #\ZERO_WIDTH_JOINER #\MICROPHONE)
             (#\MAN #\ZERO_WIDTH_JOINER #\GRADUATION_CAP)
             (#\MAN #\ZERO_WIDTH_JOINER #\BABY_BOTTLE)
             (#\MAN #\ZERO_WIDTH_JOINER #\COOKING)
             (#\MAN #\ZERO_WIDTH_JOINER #\EAR_OF_RICE)
             (#\MAN #\ZERO_WIDTH_JOINER #\AIRPLANE #\VARIATION_SELECTOR-16)
             (#\MAN #\ZERO_WIDTH_JOINER #\SCALES #\VARIATION_SELECTOR-16)
             (#\MAN #\ZERO_WIDTH_JOINER #\STAFF_OF_AESCULAPIUS
              #\VARIATION_SELECTOR-16)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\CHRISTMAS_TREE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\CHRISTMAS_TREE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\CHRISTMAS_TREE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\CHRISTMAS_TREE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\CHRISTMAS_TREE)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\U1F9D1 #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2
              #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16
              #\ZERO_WIDTH_JOINER #\KISS_MARK #\ZERO_WIDTH_JOINER #\U1F9D1
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\U1F9D1 #\ZERO_WIDTH_JOINER #\U1F91D #\ZERO_WIDTH_JOINER
              #\U1F9D1)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\WOMAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\WOMAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\GIRL)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\BOY
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\GIRL #\ZERO_WIDTH_JOINER #\GIRL)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\GIRL #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\GIRL)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\BOY #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\BOY)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\KISS_MARK
              #\ZERO_WIDTH_JOINER #\WOMAN)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\KISS_MARK
              #\ZERO_WIDTH_JOINER #\MAN)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\WOMAN)
             (#\WOMAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\MAN)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\U1F91D #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\KISS_MARK #\ZERO_WIDTH_JOINER #\MAN
              #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-6)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-5)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-4)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-3)
             (#\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2 #\ZERO_WIDTH_JOINER
              #\HEAVY_BLACK_HEART #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER
              #\MAN #\EMOJI_MODIFIER_FITZPATRICK_TYPE-1-2)
             (#\MAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\BOY
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\WOMAN #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\MAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\MAN #\ZERO_WIDTH_JOINER #\GIRL
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\MAN #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\MAN #\ZERO_WIDTH_JOINER #\BOY
              #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\MAN #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\GIRL #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\GIRL #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\GIRL)
             (#\MAN #\ZERO_WIDTH_JOINER #\BOY #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\BOY)
             (#\MAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\KISS_MARK
              #\ZERO_WIDTH_JOINER #\MAN)
             (#\MAN #\ZERO_WIDTH_JOINER #\HEAVY_BLACK_HEART
              #\VARIATION_SELECTOR-16 #\ZERO_WIDTH_JOINER #\MAN))))

(defgeneric emojip (codepoints-chars))

(defmethod emojip ((codepoints-chars list))
  (first (or (emoji-zwj-sequences-p codepoints-chars)
             (emoji-sequences-p     codepoints-chars))))

(defgeneric starting-emoji (object))

(defmethod starting-emoji ((object list))
  (emojip object))

(defmethod starting-emoji ((object string))
  (emojip (coerce object 'list)))
